## Copyright (C) 2022 Ioannis Kosmidis

## Distributed as part of the supplementary material for the manuscript
## "Empirical bias-reducing adjustments to estimating functions"
##
## Authors: Ioannis Kosmidis
## Licence: GPL 2 or greater
## NOT A POLISHED PIECE OF PUBLIC-USE SOFTWARE!  Provided "as is".
## NO WARRANTY OF FITNESS FOR ANY PURPOSE!
##
## It is assumed that the user has a Julia and R installation with all
## packages required by the current script and any other script
## included.
##
##
## Code for computing the coverage probabilities and expected lenghts
## reported in Exammple 5.3

using Random, Autologistic, StatsBase, Optim, RCall

# Subset  2
area_to_use = 1
min_village = 2
max_village = 14
include("gambia_prepare_data.jl")

n_simu = 1000
R = 500
verbose = 1
o_thres = 10000.0;

## Configure the optimizer
optim_options = Optim.Options(x_tol = 1e-12,
                              f_tol = 1e-12,
                              iterations = 1000,
                              allow_f_increases = true);

## RBM+ estimation
res_rbmp = auto_symmetric.fit(zeros(pall), gambia_data, true, subsets = subsets, plugin = true,
                              optim_options = optim_options, auto_diff = true);

## Estimates
theta = res_rbmp.minimizer


## Objects to store results
coef_m = fill(NaN, n_simu, pall);
coef_rbmp = fill(NaN, n_simu, pall);
se_m = fill(NaN, n_simu, pall);
se_rbmp = fill(NaN, n_simu, pall);
grad_m = fill(NaN, n_simu, pall);
grad_rbmp = fill(NaN, n_simu, pall);
infinite_m = fill(NaN, n_simu, pall);
converged_m = fill(false, n_simu);
converged_rbmp = fill(false, n_simu);
objective_m = fill(NaN, n_simu);
objective_rbmp = fill(NaN, n_simu);
boots_m = Vector(undef, n_simu);
boots_rbmp = Vector(undef, n_simu);

seed = 20220502;
Random.seed!(seed);;
responses = auto_symmetric.simulate(theta, gambia_data, n_simu);

# Find out which maximum pseudolikelihood estimates will be infinite
for i in 1:n_simu
    temp_data = auto_symmetric.Data(responses[:, i], X, G)
    infinite_m[i, :] = auto_symmetric.is_separated(temp_data)
    auto_symmetric.progress(i, n_simu, "Separation detection: ", verbose)
end


# M estimation
for i in 1:n_simu
    temp_data = auto_symmetric.Data(responses[:, i], X, G)
    # Try from zero
    res = auto_symmetric.get_results(temp_data, zeros(pall), false, subsets = subsets, plugin = false, optim_options = optim_options)
    # Try from truth
    if !res["converged"] | (abs(res["objective"]) > o_thres)
        res = auto_symmetric.get_results(temp_data, true_theta, false, subsets = subsets, plugin = false, optim_options = optim_options)
    end
    # Retrieve results
    converged_m[i] = res["converged"]
    coef_m[i, :] = res["coefs"]
    grad_m[i, :] = res["grad"]
    se_m[i, :] = res["ses"]
    objective_m[i] = res["objective"]
    boots_m[i] = auto_symmetric.boot(
        res["coefs"],
        temp_data,
        false,
        R,
        subsets = subsets,
        plugin = false,
        verbose = 10000,
        Mstart = false,
        optim_options = optim_options,
        alt_start = zeros(pall))
    auto_symmetric.progress(i, n_simu, "M estimation: ", verbose)
end

# RBM+ estimation
for i in 1:n_simu
    temp_data = auto_symmetric.Data(responses[:, i], X, G)
    # Try from zero
    res = auto_symmetric.get_results(temp_data, zeros(pall), true, subsets = subsets, plugin = true, optim_options = optim_options)
    # Try from M
    if !res["converged"] | (abs(res["objective"]) > o_thres)
        res = auto_symmetric.get_results(temp_data, coef_m[i, :], true, subsets = subsets, plugin = true, optim_options = optim_options)
    elseif !res["converged"] | (abs(res["objective"]) > o_thres)
        # Try from truth
        res = auto_symmetric.get_results(temp_data, true_theta, true, subsets = subsets, plugin = true, optim_options = optim_options)
    end
    # Retrieve results
    converged_rbmp[i] = res["converged"]
    coef_rbmp[i, :] = res["coefs"]
    grad_rbmp[i, :] = res["grad"]
    se_rbmp[i, :] = res["ses"]
    objective_rbmp[i] = res["objective"]
    boots_rbmp[i] = auto_symmetric.boot(
        res["coefs"],
        temp_data,
        true,
        R,
        subsets = subsets,
        plugin = true,
        verbose = 10000,
        Mstart = false,
        optim_options = optim_options,
        alt_start = zeros(pall))
    auto_symmetric.progress(i, n_simu, "RBM+plugin estimation: ", verbose)
end




## Relative bias
(mean.(eachcol(coef_m)) - theta)./theta
(mean.(eachcol(coef_rbmp)) - theta)./theta

## Bias over se (See Efron adn Tibshirani)
(mean.(eachcol(coef_m)) - theta)./ sqrt.(var.(eachcol(coef_m)))
(mean.(eachcol(coef_rbmp)) - theta)./ sqrt.(var.(eachcol(coef_rbmp)))

## Bias^2 over var
(mean.(eachcol(coef_m)) - theta).^2 ./ var.(eachcol(coef_m))
(mean.(eachcol(coef_rbmp)) - theta).^2 ./ var.(eachcol(coef_rbmp))

summaries_m = Vector{Dict}(undef, n_simu)
summaries_rbmp = Vector{Dict}(undef, n_simu)
for k in 1:n_simu
    auto_symmetric.progress(k, n_simu, "sample", 1)
    # if all(any.(eachrow(boots_m[k]["is_infinite"])))
    #     summaries_m[k] = Dict()
    # else
        summaries_m[k] = auto_symmetric.summarize_boot(boots_m[k], C = Inf, alpha = 0.05)
    # end
   summaries_rbmp[k] = auto_symmetric.summarize_boot(boots_rbmp[k], C = Inf, alpha = 0.05)
end

## Coverage
cover_rbmp_perc = Matrix(undef, n_simu, pall);
cover_rbmp_basic = Matrix(undef, n_simu, pall);
cover_rbmp_norm = Matrix(undef, n_simu, pall);
for k in 1:n_simu 
    cover_rbmp_norm[k, :] = (summaries_rbmp[k]["ci_norm"][:, 1] .< theta) .& (summaries_rbmp[k]["ci_norm"][:, 2] .> theta)
    cover_rbmp_perc[k, :] = (summaries_rbmp[k]["ci_perc"][:, 1] .< theta) .& (summaries_rbmp[k]["ci_perc"][:, 2] .> theta)
    cover_rbmp_basic[k, :] = (summaries_rbmp[k]["ci_basic"][:, 1] .< theta) .& (summaries_rbmp[k]["ci_basic"][:, 2] .> theta)
end

## Length
length_rbmp_perc = Matrix(undef, n_simu, pall);
length_rbmp_basic = Matrix(undef, n_simu, pall);
length_rbmp_norm = Matrix(undef, n_simu, pall);
for k in 1:n_simu 
    length_rbmp_norm[k, :] = summaries_rbmp[k]["ci_norm"][:, 2] - summaries_rbmp[k]["ci_norm"][:, 1]
    length_rbmp_perc[k, :] = summaries_rbmp[k]["ci_perc"][:, 2] - summaries_rbmp[k]["ci_perc"][:, 1]
    length_rbmp_basic[k, :] = summaries_rbmp[k]["ci_basic"][:, 2] - summaries_rbmp[k]["ci_basic"][:, 1]
end

## Coverage
cover_m_perc = Matrix(undef, n_simu, 7);
for k in 1:n_simu
    if length(summaries_m[k]) == 0
        continue
    end
    cover_m_perc[k, :] = (summaries_m[k]["ci_perc"][:, 1] .< theta) .& (summaries_m[k]["ci_perc"][:, 2] .> theta)
    println(k)
end

## Length
length_m_perc = Matrix(undef, n_simu, pall);
for k in 1:n_simu
    println(k)
    if !inds[k]
        length_m_perc[k, :] = summaries_m[k]["ci_perc"][:, 2] - summaries_m[k]["ci_perc"][:, 1]
    end
end


inds = map(x -> length(summaries_m[x]) == 0, 1:n_simu)
## Coverage
hcat(round.(mean(eachrow(cover_m_perc[.!inds, :])), digits = 3), round.(mean(eachrow(cover_rbmp_perc)), digits = 3))
## Lengths
al_m_perc = mean(eachrow(length_m_perc[.!inds, :]))
al_rbmp_perc = mean(eachrow(length_rbmp_perc))
hcat(round.(al_m_perc, digits = 3), round.(al_rbmp_perc, digits = 3))


objects = ["X", "Y", "elist", "df", "responses",  "theta", "pall", "n_simu",
           "parnames", "subsets", 
           "coef_m", "coef_rbmp",
           "infinite_m",
           "se_m", "se_rbmp",
           "grad_m", "grad_rbmp",
           "boots_m", "boots_rbmp",
           "summaries_m", "summaries_rbmp",
           "converged_m", "converged_rbmp",
           "objective_m", "objective_rbmp",           
           "seed"]

file_name = "gambia_simulation_boot.rda"
for i in eachindex(objects)
        current_object = objects[i]
        val = eval(Meta.parse(current_object))
        R"""
           assign($current_object, $val)
        """
end

# save
R"""
save(list = $objects, file = $file_name)
"""
