## Copyright (C) 2022 Ioannis Kosmidis

## Distributed as part of the supplementary material for the manuscript
## "Empirical bias-reducing adjustments to estimating functions"
##
## Authors: Ioannis Kosmidis
## Licence: GPL 2 or greater
## NOT A POLISHED PIECE OF PUBLIC-USE SOFTWARE!  Provided "as is".
## NO WARRANTY OF FITNESS FOR ANY PURPOSE!
##
## It is assumed that the user has a Julia and R installation with all
## packages required by the current script and any other script
## included.
##
##
## Code used by the scripts that reproduce the results in Example 5.3 and 5.5

## Penalized likelihood
loglik <- function(pars, X, y, link, penalize = TRUE, penalty = "PML",
                   scores_fun, info_fun) {
    probs <- link$linkinv(drop(X %*% pars))
    loglik <- sum(dbinom(y, 1, probs, log = TRUE))
    n <- length(y)
    if (penalize) {
        j <- info_fun(pars, type = "observed")
        ii <- info_fun(pars, type = "expected")
        so <- crossprod(scores_fun(pars, contributions = TRUE))
        pen <- try(switch(penalty,
                          "PML" = 0.5 * log(det(ii)),
                          "det1" = 0.5 * log(det(j)) - 0.5 * log(det(so)),
                          "RBM" = - 0.5 * sum(diag(solve(j) %*% so)),
                          "RBM_extra" = - 0.5 * sum(diag(solve(j) %*% so)) + log(det(ii)) / n),
                   silent = TRUE)
        if (inherits(pen, "try-error")) {
            out <- NA
            attr(out, "penalty") <- NA
            return(out)
        }
        else {
            out <- loglik + pen
            attr(out, "penalty") <- pen
        }
    }
    else {
        out <- loglik
            attr(out, "penalty") <- 0
    }
    attr(out, "loglik") <- loglik
    out
}

fit_model <- function(X, y, link, silent = TRUE) {
    ## ML
    fit_ml <- glm(y ~ - 1 + X, family = binomial(link), epsilon = 1e-12, maxit = 1000)
    info_fun <- get_information_function(fit_ml)
    scores_fun <- get_score_function(fit_ml)
    scores_ml <- scores_fun(coef(fit_ml))
    coefs_ml <- coef(fit_ml)
    converged_ml <- all(abs(scores_ml) < 1e-03)
    n_p <- length(coefs_ml)

    ## BR
    fit_br <- glm(y ~ - 1 + X, family = binomial(link), epsilon = 1e-08, maxit = 1000,
                  method = "brglm_fit", type = "AS_mean")
    coefs_br <- coef(fit_br)
    scores_br <- fit_br$grad[1:n_p]
    converged_br <- all(abs(scores_br) < 1e-03)

    ## RBM
    fit_tr <- optimx(coefs_br,
                     fn = function(pars) {
        out <- -loglik(pars, X, y, link, TRUE, "RBM", scores_fun, info_fun)
        attributes(out) <- NULL
        out
    }, method = c("BFGS", "nlminb"))
    out_tr <- summary(fit_tr, order = "value")
    coefs_tr <- coef(out_tr)[1, ]
    scores_tr <- attr(out_tr, "details")[, "ngatend"][[1]]
    converged_tr <- all(abs(scores_tr) < 1e-03)

    ## RBMp
    fit_tr_extra <- optimx(coefs_br,
                     fn = function(pars) {
        out <- -loglik(pars, X, y, link, TRUE, "RBM_extra", scores_fun, info_fun)
        attributes(out) <- NULL
        out
    }, method = c("BFGS", "nlminb"))
    out_tr_extra <- summary(fit_tr_extra, order = "value")
    coefs_tr_extra <- coef(out_tr_extra)[1, ]
    scores_tr_extra <- attr(out_tr_extra, "details")[, "ngatend"][[1]]
    converged_tr_extra <- all(abs(scores_tr_extra) < 1e-03)

    ## eRBM
    rbm_penalty <- function(pars) {
        attr(loglik(pars, X, y, link, TRUE, "RBM", scores_fun, info_fun), "penalty")
    }
    jinv <- solve(info_fun(coefs_ml, type = "observed"))
    coefs_eRBM <- try(drop(coefs_ml + jinv %*% numDeriv::grad(rbm_penalty, x = coefs_ml)), silent = TRUE)
    if (inherits(coefs_eRBM, "try-error"))
        coefs_eRBM <- rep(NA, n_p)

    ## Separation detection
    sepa <- try(glm(y ~ -1 + X, family = binomial(link), method = "detect_separation"), silent = silent)
    if (inherits(sepa, "try-error")) {
        separated <-  NA
    }
    else {
        separated <- sepa$outcome
        infinite_ml <- coef(sepa)
    }

    list(coef_ml = coefs_ml,
         coef_tr = coefs_tr,
         coef_br = coefs_br,
         coef_tr_extra = coefs_tr_extra,
         coef_erbm = coefs_eRBM,
         scores_ml = scores_ml,
         scores_tr = scores_tr,
         scores_br = scores_br,
         scores_tr_extra = scores_tr_extra,
         scores_erbm = rep(NA, n_p),
         separated = separated,
         converged_tr = converged_tr,
         converged_tr_extra = converged_tr_extra,
         converged_br = converged_br,
         converged_ml = converged_ml,
         converged_eRBM = converged_ml,
         infinite_ml = infinite_ml)
}

## Assuming the the first column is ones!
## start has length ncol(X)
fit_all_models <- function(X, y, link, silent = TRUE) {
    p <- ncol(X) - 1
    inout <- list(0:1)
    for (j in 2:p) {
        inout <- c(inout, list(0:1))
    }
    inout <- cbind(1, expand.grid(inout))
    n_obs <- length(y)
    n_models <- nrow(inout)
    coefs_ml <- coefs_tr <- coefs_tr_extra <- coefs_br <- infinite_ml <- matrix(NA, n_models, ncol(X))
    ic_names <- c("tic", "aic", "bic")
    ic_tr <- ic_tr_extra <- ic_ml <- ic_br <- matrix(NA, n_models, length(ic_names))
    model_names <- apply(inout, 1, function(x) paste(which(x == 1), collapse = "-"))
    separated <- converged_tr <- converged_tr_extra <- converged_ml <- converged_br <- logical(n_models)
    colnames(inout) <- colnames(X)
    colnames(ic_ml) <- colnames(ic_tr) <- colnames(ic_br) <- colnames(ic_tr_extra) <- ic_names
    rownames(inout) <- rownames(ic_ml) <- rownames(ic_tr) <- rownames(ic_br) <- rownames(ic_tr_extra) <-
        rownames(coefs_ml) <- rownames(coefs_tr) <- rownames(coefs_br) <- rownames(coefs_tr_extra) <-
        rownames(infinite_ml) <-
        names(separated) <-
        names(converged_ml) <- names(converged_br) <- names(converged_tr) <- names(converged_tr_extra) <-
        model_names
    for (j in seq.int(n_models)) {
        which_cov <- which(inout[j, ] == 1)
        X_sub <- X[, which_cov, drop = FALSE]
        ml <- glm(y ~ - 1 + X_sub, family = binomial(link), epsilon = 1e-12, maxit = 1000)
        br <- glm(y ~ - 1 + X_sub, family = binomial(link), epsilon = 1e-08, maxit = 1000,
                  method = "brglm_fit", type = "AS_mean")
        sepa <- try(glm(y ~ -1 + X_sub, family = binomial(link), method = "detect_separation"), silent = silent)
        if (inherits(sepa, "try-error")) {
            separated[j] <-  NA
        }
        else {
            separated[j] <- sepa$outcome
            infinite_ml[j, which_cov] <- coef(sepa)
        }
        info_fun <- get_information_function(ml)
        scores_fun <- get_score_function(ml)
        n_p <- length(which_cov)
        if (isTRUE(n_p == 1)) {
            summ <- coef(summary(ml))
            lims <- summ[, 1] + c(-1, 1) * 5 * summ[, 2]
        }
        else {
            lims <- c(-1, 1) * Inf
        }
        fit_tr <- try(nlm(function(par, ...) -loglik(par, ...), p = coef(br), X = X_sub, y = y, link = link,
                      penalize = TRUE, penalty = "RBM",
                      scores = scores_fun, info = info_fun), silent = silent)
        if (inherits(fit_tr, "try-error")) {
            for (m in c("BFGS", "Nelder-Mead")) {
                fit_tr <- try(optim(coef(br),
                                    fn = loglik, X = X_sub, y = y, link = link,
                                    penalize = TRUE, penalty = "RBM",
                                    scores = scores_fun, info = info_fun,
                                    method = if (isTRUE(n_p == 1)) "Brent" else m,
                                    lower = lims[1],
                                    upper = lims[2],
                                    control = list(fnscale = -1, reltol = 1e-15, maxit = 1000)), silent = silent)
                if (!inherits(fit_tr, "try-error")) {
                    break
                }
            }
            scores_tr <- try(numDeriv::grad(loglik, fit_tr$par, X = X_sub, y = y, link = link,
                                            penalize = TRUE, penalty = "RBM",
                                            scores = scores_fun, info = info_fun), silent = silent)
            if (is(scores_tr, "try-error"))
                scores_tr <- rep(NA, ncol(X_sub))
        }
        else {
            fit_tr$par <- fit_tr$estimate
            fit_tr$value <- fit_tr$minimum
            scores_tr <- fit_tr$gradient
        }
        fit_tr_extra <- try(nlm(function(par, ...) -loglik(par, ...), p = coef(br),
                                X = X_sub, y = y, link = link,
                                penalize = TRUE, penalty = "RBM_extra",
                                scores = scores_fun, info = info_fun), silent = silent)
        if (inherits(fit_tr_extra, "try-error")) {
            for (m in c("BFGS", "Nelder-Mead")) {
                fit_tr_extra <- try(optim(coef(br),
                                          fn = loglik, X = X_sub, y = y, link = link,
                                          penalize = TRUE, penalty = "RBM_extra",
                                          scores = scores_fun, info = info_fun,
                                          method = if (isTRUE(n_p == 1)) "Brent" else m,
                                          lower = lims[1],
                                          upper = lims[2],
                                          control = list(fnscale = -1, reltol = 1e-15, maxit = 1000)), silent = silent)
                if (!inherits(fit_tr_extra, "try-error")) {
                    break
                }
            }
            scores_tr_extra <- try(numDeriv::grad(loglik, fit_tr_extra$par, X = X_sub, y = y,
                                                  link = link,
                                                  penalize = TRUE, penalty = "RBM_extra",
                                                  scores = scores_fun, info = info_fun), silent = silent)
            if (is(scores_tr_extra, "try-error"))
                scores_tr <- rep(NA, ncol(X_sub))
            }
        else {
            fit_tr_extra$par <- fit_tr_extra$estimate
            fit_tr_extra$value <- fit_tr_extra$minimum
            scores_tr_extra <- fit_tr_extra$gradient
        }
        scores_ml <- scores_fun(coef(ml))
        converged_tr[j] <- all(abs(scores_tr) < 1e-03)
        converged_tr_extra[j] <- all(abs(scores_tr_extra) < 1e-03)
        converged_br[j] <- all(abs(br$grad[1:n_p]) < 1e-03)
        converged_ml[j] <- all(abs(scores_ml) < 1e-03)
        cc_ml <- coef(ml)
        cc_tr <- fit_tr$par
        cc_tr_extra <- fit_tr_extra$par
        coefs_ml[j, which_cov] <- cc_ml
        coefs_br[j, which_cov] <- coef(br)
        coefs_tr[j, which_cov] <- cc_tr
        coefs_tr_extra[j, which_cov] <- cc_tr_extra
        ll_ml <- loglik(coefs_ml[j, which_cov], X_sub, y, link, penalize = TRUE, penalty = "RBM",
                        scores = scores_fun, info = info_fun)
        ll_tr <- loglik(coefs_tr[j, which_cov], X_sub, y, link, penalize = TRUE, penalty = "RBM",
                        scores = scores_fun, info = info_fun)
        ll_br <- loglik(coefs_br[j, which_cov], X_sub, y, link, penalize = TRUE, penalty = "RBM",
                        scores = scores_fun, info = info_fun)
        ll_tr_extra <- loglik(coefs_tr_extra[j, which_cov], X_sub, y, link, penalize = TRUE, penalty = "RBM",
                              scores = scores_fun, info = info_fun)
        ic_ml[j, ] <- c(tic = -2 * ll_ml - 2 * attr(ll_ml, "penalty"),
                        aic = -2 * ll_ml + 2 * n_p,
                        bic = -2 * ll_ml + n_p * log(n_obs))
        ic_br[j, ] <- c(tic = -2 * ll_br - 2 * attr(ll_br, "penalty"),
                        aic = -2 * ll_br + 2 * n_p,
                        bic = -2 * ll_br + n_p * log(n_obs))
        ic_tr[j, ] <- c(tic = -2 * ll_tr - 2 * attr(ll_tr, "penalty"),
                        aic = -2 * ll_tr + 2 * n_p,
                        bic = -2 * ll_tr + n_p * log(n_obs))
        ic_tr_extra[j, ] <- c(tic = -2 * ll_tr_extra - 2 * attr(ll_tr_extra, "penalty"),
                              aic = -2 * ll_tr_extra + 2 * n_p,
                              bic = -2 * ll_tr_extra + n_p * log(n_obs))
    }
    list(coef_ml = coefs_ml,
         coef_tr = coefs_tr,
         coef_br = coefs_br,
         coef_tr_extra = coefs_tr_extra,
         ic_ml = ic_ml,
         ic_tr = ic_tr,
         ic_tr_extra = ic_tr_extra,
         ic_br = ic_br,
         separated = separated,
         converged_tr = converged_tr,
         converged_tr_extra = converged_tr_extra,
         converged_br = converged_br,
         converged_ml = converged_ml,
         infinite_ml = infinite_ml,
         models = inout)
}
